############################################################################################
# Demo of training a deep neural network on the MNIST handwriting digit recognition dataset.
#
# Dependency: require(RevoScaleR)
# Note: internet access required for initial download of data from
# http://yann.lecun.com/exdb/mnist/
############################################################################################

library("MicrosoftML")

# This file contains the Net# definition for the neural network structure.
# More information about the Net# language here: 
# https://azure.microsoft.com/en-us/documentation/articles/machine-learning-azure-ml-netsharp-reference-guide/
netDefFile <- system.file("demoScripts/mnist.nn", package = "MicrosoftML")

source(system.file("extdata/mnist.R", package = "MicrosoftML"))

mnist <- getMnistData(download = TRUE, sampleDataDir = NULL, createDir = TRUE)
mnistTrain <- mnist$mnistTrain
mnistTest <- mnist$mnistTest
rm(mnist)

rxGetInfo(mnistTrain)

# multiClass with rxNeuralNet
nnModel <- rxNeuralNet(Label ~ .,
    data = mnistTrain,
    numIterations = 10,
    normalize = "no",
    optimizer = sgd(learningRate=0.05, lRateRedRatio=0.9, lRateRedFreq=5),
    netDefinition = readChar(netDefFile, file.info(netDefFile)$size),
    type = "multi")

# Runs the prediction model on the test data frame.
nnScore <- rxPredict(nnModel, mnistTest, extraVarsToWrite = "Label")
rxGetVarInfo(nnScore)

# Evaluate the performance
rxCrossTabs(formula = ~ F(Label):PredictedLabel, data = nnScore)

# Show the (micro-)accuracy
sum(nnScore$Label == nnScore$PredictedLabel)/dim(nnScore)[1]

# multiClass with rxLogisticRegression
logitModel <- rxLogisticRegression(Label ~ ., 
    type = "multiClass",
    data = mnistTrain)
summary(logitModel)
logitScore <- rxPredict(logitModel, mnistTest, extraVarsToWrite = "Label")
rxGetVarInfo(logitScore)
rxCrossTabs(formula = ~ F(Label):PredictedLabel, data = logitScore)
sum(logitScore$Label == logitScore$PredictedLabel)/dim(logitScore)[1]