#############################################################################
# Regression examples using the communities data
#
# Dependency: require(RevoScaleR)
# Note: internet access required for initial download of data from
# https://archive.ics.uci.edu/ml/machine-learning-databases/communities/
#############################################################################

library("MicrosoftML")

# Define RMSE which will be used as performance metric
rmse <- function(actual, estimated)
{
    return(sqrt(mean((actual-estimated)^2, na.rm=TRUE)))
}

source(system.file("extdata/communities.R", package = "MicrosoftML"))

communities <- getCommunitiesData(download = TRUE, sampleDataDir = NULL, createDir = TRUE)
communitiesTrain <- communities$communitiesTrain
communitiesTest <- communities$communitiesTest
rm(communities)

communitiesFormula <- ViolentCrimesPerPop ~ .

# Regression Fast Tree
treeModel <- rxFastTrees(communitiesFormula, type = "regression",
    data = communitiesTrain,
    mlTransforms = categorical(vars = "communityname"))
summary(treeModel)
treeScore <- rxPredict(treeModel, data = communitiesTest,
    extraVarsToWrite = c("ViolentCrimesPerPop"))
rmse(treeScore$ViolentCrimesPerPop, treeScore$Score)

# Regression Fast Forest
forestModel <- rxFastForest(communitiesFormula, type = "regression",
    data = communitiesTrain,
    mlTransforms = categorical(vars = "communityname"),
    mlTransformVars = list("communityname"))
summary(forestModel)
forestScore <- rxPredict(forestModel, data = communitiesTest,
    extraVarsToWrite = c("ViolentCrimesPerPop"))
rmse(forestScore$ViolentCrimesPerPop, forestScore$Score)

# Regression Neural Net
nnModel <- rxNeuralNet(communitiesFormula, type = "regression",
    data = communitiesTrain,
    mlTransforms = categorical(vars = "communityname"))
nnScore <- rxPredict(nnModel, data = communitiesTest,
    extraVarsToWrite = c("ViolentCrimesPerPop"))
rmse(nnScore$ViolentCrimesPerPop, nnScore$Score)