## test low level functions generated from entrypoints but not exported.

"test.ImportTextData.ImportText.TextLoader" <- function()
{
    ## use a local function to wrap the non-exported functions
    func <- function(inputFile, customSchema, predictorModel)
    {
        importNode <- ImportTextData.ImportText(
            InputFile = "$inputFile"
            , Data = "$importedData"
            , CustomSchema = customSchema
            )
        normalizeNode <- Normalize.MinMax(
            Data = "$importedData"
            #, Columns = "Features"
            , Column = list(list(Source = "Features"))
            , OutputData = "$normalizedData"
            )
        trainNode <- LogisticRegression.TrainBinary(
            TrainingData = "$normalizedData"
            , PredictorModel = "$predictorModel"
            , NumThreads = 4
            )

        exobj <- epCreateGraph(
            importNode, normalizeNode, trainNode
            , inputs = list(inputFile = inputFile)
            , outputs = list(predictorModel = predictorModel)
            )

        ## make sure all the methods for epRunGraph work
        ## only the result of the last one is returned.
        exjson <- epRunGraph(unclass(toJSON(exobj)))
        exjson <- epRunGraph(toJSON(exobj))
        exjson <- epRunGraph(exobj)
    }

    ## set the function environment so that non-exported functions can be found
    environment(func) <- asNamespace("MicrosoftML")

    ## prepare data
    binaryDf <- within(iris, isSetosa <- Species == "setosa")
    colInfo <- rxCreateColInfo(binaryDf)
    binaryFile <- tempfile("binaryFile", fileext = ".txt")
    binaryText <- RxTextData(binaryFile, colInfo = colInfo)
    rxDataStep(binaryDf, binaryText, overwrite = TRUE)

    #file.show(binaryFile)
    customSchema = "separator=, hasHeader=+ column=Label:BL:5 column=Features:R4:0-3"

    ## get model in memory
    modelInMemory <- try(func(inputFile = binaryFile, customSchema = customSchema, predictorModel = ""))
    checkTrue(!inherits(modelInMemory, "try-error"))
   
    ## write model to disk
    predModel <- tempfile("predModel", fileext = ".zip")
    on.exit(if(file.exists(predModel)) unlink(predModel), add = TRUE)
    modelOnDisk <- try(func(inputFile = binaryFile, customSchema = customSchema, predictorModel = predModel))
    checkTrue(!inherits(modelOnDisk, "try-error"))

    ## check
    #modelFileSize <- file.info(predModel)$size
    #checkEquals(modelFileSize, length(modelInMemory$outModelBin))

    ##TODO: why are the bytes not always the same?
    #outModelBin <- readBin(predModel, what = "raw", n = modelFileSize)
    #checkEquals(outModelBin, modelInMemory$outModelBin)
}

"test.FeatureCombiner.PrepareFeatures" <- function()
{
    func <- function(inputData, label, features, predictorModel)
    {
        featureNode <- FeatureCombiner.PrepareFeatures(
            Data = "$inputData"
            , Features = features
            , OutputData = "$featureData"
            )
        trainNode <- LogisticRegression.TrainBinary(
            TrainingData = "$featureData"
            , PredictorModel = "$predictorModel"
            , LabelColumn = label
            , FeatureColumn = "Features"
            )

        exobj <- epCreateGraph(
            featureNode, trainNode
            , inputs = list(inputData = "")
            , outputs = list(predictorModel = predictorModel)
            )
        exjson <- epRunGraph(exobj, data = inputData)
    }

    environment(func) <- asNamespace("MicrosoftML")

    ###################################
    ## data.frame
    binaryDf <- within(iris, isSetosa <- Species == "setosa")
    label <- "isSetosa"
    features <- c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")

    predModel <- tempfile("predModel", fileext = ".zip")
    on.exit(if(file.exists(predModel)) unlink(predModel), add = TRUE)

    modelDfOnDisk <- try(func(inputData = binaryDf, label = label, features = features, predictorModel = predModel))
    checkTrue(!inherits(modelDfOnDisk, "try-error"))

    modelDfInMemory <- try(func(inputData = binaryDf, label = label, features = features, predictorModel = ""))
    checkTrue(!inherits(modelDfInMemory, "try-error"))

    ## check
    #modelFileSize <- file.info(predModel)$size
    #checkEquals(modelFileSize, length(modelDfInMemory$outModelBin))

    #outModelBin <- readBin(predModel, what = "raw", n = modelFileSize)
    #checkEquals(outModelBin, modelDfInMemory$outModelBin)

    ###################################
    ## xdf
    binaryXdf <- tempfile("binaryXdf", fileext = ".xdf")
    rxDataStep(binaryDf, binaryXdf, overwrite = TRUE)
    on.exit(if(file.exists(binaryXdf)) unlink(binaryXdf), add = TRUE)

    modelXdfOnDisk <- try(func(inputData = binaryXdf, label = label, features = features, predictorModel = predModel))
    checkTrue(!inherits(modelXdfOnDisk, "try-error"))

    modelXdfInMemory <- try(func(inputData = binaryXdf, label = label, features = features, predictorModel = ""))
    checkTrue(!inherits(modelXdfInMemory, "try-error"))

    ## check
    #modelFileSize <- file.info(predModel)$size
    #checkEquals(modelFileSize, length(modelXdfInMemory$outModelBin))

    #outModelBin <- readBin(predModel, what = "raw", n = modelFileSize)
    #checkEquals(outModelBin, modelXdfInMemory$outModelBin)

    ###################################
    ## txt
    colInfo <- rxCreateColInfo(binaryDf)
    binaryFile <- tempfile("binaryFile", fileext = ".txt")
    binaryText <- RxTextData(binaryFile, colInfo = colInfo)

    rxDataStep(binaryDf, binaryText)
    on.exit(if(file.exists(binaryFile)) unlink(binaryFile), add = TRUE)

    modelTxtOnDisk <- try(func(inputData = binaryText, label = label, features = features, predictorModel = predModel))
    checkTrue(!inherits(modelTxtOnDisk, "try-error"))

    modelTxtInMemory <- try(func(inputData = binaryText, label = label, features = features, predictorModel = ""))
    checkTrue(!inherits(modelTxtInMemory, "try-error"))

    ## check
    #modelFileSize <- file.info(predModel)$size
    #checkEquals(modelFileSize, length(modelTxtInMemory$outModelBin))

    #outModelBin <- readBin(predModel, what = "raw", n = modelFileSize)
    #checkEquals(outModelBin, modelTxtInMemory$outModelBin)
}

"test.Normalize.MeanVar" <- function()
{
    func <- function(inputData, label, features, predictorModel)
    {
        featureNode <- FeatureCombiner.PrepareFeatures(
            Data = "$inputData"
            , Features = features
            , OutputData = "$featureData"
            )
        normalizeNode <- Normalize.MeanVar(
            Data = "$featureData"
            #, Columns = "Features"
            , Column = list(list(Source = "Features"))
            , OutputData = "$normalizedData"
            )
        trainNode <- LogisticRegression.TrainBinary(
            TrainingData = "$normalizedData"
            , PredictorModel = "$predictorModel"
            , LabelColumn = label
            , FeatureColumn = "Features"
            )

        exobj <- epCreateGraph(
            featureNode, normalizeNode, trainNode
            , inputs = list(inputData = "")
            , outputs = list(predictorModel = predictorModel)
            )
        exjson <- epRunGraph(exobj, data = inputData)
    }

    environment(func) <- asNamespace("MicrosoftML")

    ###################################
    ## data.frame
    binaryDf <- within(iris, isSetosa <- Species == "setosa")
    label <- "isSetosa"
    features <- c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")

    predModel <- tempfile("predModel", fileext = ".zip")
    on.exit(if(file.exists(predModel)) unlink(predModel), add = TRUE)

    modelDfOnDisk <- try(func(inputData = binaryDf, label = label, features = features, predictorModel = predModel))
    checkTrue(!inherits(modelDfOnDisk, "try-error"))

    modelDfInMemory <- try(func(inputData = binaryDf, label = label, features = features, predictorModel = ""))
    checkTrue(!inherits(modelDfInMemory, "try-error"))

    ## check
    #modelFileSize <- file.info(predModel)$size
    #checkEquals(modelFileSize, length(modelDfInMemory$outModelBin))

    #outModelBin <- readBin(predModel, what = "raw", n = modelFileSize)
    #checkEquals(outModelBin, modelDfInMemory$outModelBin)
}

"test.ScoreModel.MakeScoringTransform" <- function()
{
    func <- function(inputData, label, features, predictorModel)
    {
        featureNode <- FeatureCombiner.PrepareFeatures(
            Data = "$inputData"
            , Features = features
            , OutputData = "$featureData"
            )
        trainNode <- LogisticRegression.TrainBinary(
            TrainingData = "$featureData"
            , PredictorModel = "$predModel"
            , LabelColumn = label
            , FeatureColumn = "Features"
            )
        transformNode <- ScoreModel.MakeScoringTransform(
            PredictorModel = "$predModel"
            , ScoringTransform = "$predictorModel"
            )

        exobj <- epCreateGraph(
            featureNode, trainNode, transformNode
            , inputs = list(inputData = "")
            , outputs = list(predictorModel = predictorModel)
            )
        exjson <- epRunGraph(exobj, data = inputData)
    }

    environment(func) <- asNamespace("MicrosoftML")

    ###################################
    ## data.frame
    binaryDf <- within(iris, isSetosa <- Species == "setosa")
    label <- "isSetosa"
    features <- c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")

    predModel <- tempfile("predModel", fileext = ".zip")
    on.exit(if(file.exists(predModel)) unlink(predModel), add = TRUE)

    modelDfOnDisk <- try(func(inputData = binaryDf, label = label, features = features, predictorModel = predModel))
    checkTrue(!inherits(modelDfOnDisk, "try-error"))

    modelDfInMemory <- try(func(inputData = binaryDf, label = label, features = features, predictorModel = ""))
    checkTrue(!inherits(modelDfInMemory, "try-error"))

    ## check
    #modelFileSize <- file.info(predModel)$size
    #checkEquals(modelFileSize, length(modelDfInMemory$outModelBin))

    #outModelBin <- readBin(predModel, what = "raw", n = modelFileSize)
    #checkEquals(outModelBin, modelDfInMemory$outModelBin)
}

"test.TextAnalytics.TextTransformDictionary" <- function()
{
    func <- function(inputData, label, features, predictorModel)
    {
        featureNode <- TextAnalytics.TextTransform(
            Data = "$inputData"
            #, Column = features
            #, OutputColumn = "Features"
            , Column = list(Source = list(features), Name = "Features")
            , WordFeatureExtractor = NgramExtractor.NGram(MaxNumTerms = 10000000)
            , OutputData = "$featureData"
            )
        trainNode <- LogisticRegression.TrainBinary(
            TrainingData = "$featureData"
            , PredictorModel = "$predictorModel"
            , LabelColumn = label
            , FeatureColumn = "Features"
            )

        exobj <- epCreateGraph(
            featureNode, trainNode
            , inputs = list(inputData = "")
            , outputs = list(predictorModel = predictorModel)
            )
        exjson <- epRunGraph(exobj, data = inputData)
    }

    environment(func) <- asNamespace("MicrosoftML")

    ###################################
    ## data.frame
    binaryDf <- data.frame(opinion = c(
        "I love it",
        "I love it",
        "Love it",
        "I love it a lot",
        "Really love it",
        "I love it",
        "I love it",
        "love it",
        "I love it a lot",
        "Really love it",
        "I hate it",
        "I hate it",
        "I hate it",
        "hate it",
        "hate and",
        "I hate it and",
        "I hate it and",
        "I hate it and",
        "hate it and",
        "hate and"),
        like = rep(c(TRUE, FALSE), each = 10),
        stringsAsFactors = FALSE)
    label <- "like"
    features <- "opinion"

    predModel <- tempfile("predModel", fileext = ".zip")
    on.exit(if(file.exists(predModel)) unlink(predModel), add = TRUE)

    modelDfOnDisk <- try(func(inputData = binaryDf, label = label, features = features, predictorModel = predModel))
    checkTrue(!inherits(modelDfOnDisk, "try-error"))

    modelDfInMemory <- try(func(inputData = binaryDf, label = label, features = features, predictorModel = ""))
    checkTrue(!inherits(modelDfInMemory, "try-error"))

    ## check
    #modelFileSize <- file.info(predModel)$size
    #checkEquals(modelFileSize, length(modelDfInMemory$outModelBin))

    #outModelBin <- readBin(predModel, what = "raw", n = modelFileSize)
    #checkEquals(outModelBin, modelDfInMemory$outModelBin)
}

"test.TextAnalytics.TextTransformHashing" <- function()
{
    func <- function(inputData, label, features, predictorModel)
    {
        featureNode <- TextAnalytics.TextTransform(
            Data = "$inputData"
            #, Column = features
            #, OutputColumn = "Features"
            , Column = list(Source = list(features), Name = "Features")
            , CharFeatureExtractor = NgramExtractor.NGramHash(HashBits = 4)
            , OutputData = "$featureData"
            )
        trainNode <- LogisticRegression.TrainBinary(
            TrainingData = "$featureData"
            , PredictorModel = "$predictorModel"
            , LabelColumn = label
            , FeatureColumn = "Features"
            )

        exobj <- epCreateGraph(
            featureNode, trainNode
            , inputs = list(inputData = "")
            , outputs = list(predictorModel = predictorModel)
            )
        exjson <- epRunGraph(exobj, data = inputData)
    }

    environment(func) <- asNamespace("MicrosoftML")

    ###################################
    ## data.frame
    binaryDf <- data.frame(opinion = c(
        "I love it",
        "I love it",
        "Love it",
        "I love it a lot",
        "Really love it",
        "I love it",
        "I love it",
        "love it",
        "I love it a lot",
        "Really love it",
        "I hate it",
        "I hate it",
        "I hate it",
        "hate it",
        "hate and",
        "I hate it and",
        "I hate it and",
        "I hate it and",
        "hate it and",
        "hate and"),
        like = rep(c(TRUE, FALSE), each = 10),
        stringsAsFactors = FALSE)
    label <- "like"
    features <- "opinion"

    predModel <- tempfile("predModel", fileext = ".zip")
    on.exit(if(file.exists(predModel)) unlink(predModel), add = TRUE)

    modelDfOnDisk <- try(func(inputData = binaryDf, label = label, features = features, predictorModel = predModel))
    checkTrue(!inherits(modelDfOnDisk, "try-error"))

    modelDfInMemory <- try(func(inputData = binaryDf, label = label, features = features, predictorModel = ""))
    checkTrue(!inherits(modelDfInMemory, "try-error"))

    ## check
    #modelFileSize <- file.info(predModel)$size
    #checkEquals(modelFileSize, length(modelDfInMemory$outModelBin))

    #outModelBin <- readBin(predModel, what = "raw", n = modelFileSize)
    #checkEquals(outModelBin, modelDfInMemory$outModelBin)
}

"test.ModelOperations.CombineModels" <- function()
{
    func <- function(inputData, label, features, predictorModel)
    {
        featureNode <- FeatureCombiner.PrepareFeatures(
            Data = "$inputData"
            , Features = features
            , OutputData = "$featureData"
            , Model = "$dataModel"
            )
        trainNode <- LogisticRegression.TrainBinary(
            TrainingData = "$featureData"
            , PredictorModel = "$predModel"
            , LabelColumn = label
            , FeatureColumn = "Features"
            )
        combNode <- ModelOperations.CombineModels( #Error: *** Exception: 'Unsupported input type: TransformModel'
            TransformModels = "$dataModel"
        #combNode <- ModelOperations.CombineTwoModels(
            #TransformModel = "$dataModel"
            ,  PredictorModel = "$predModel"
            ,  Model = "$predictorModel")

        exobj <- epCreateGraph(
            featureNode, trainNode, combNode
            , inputs = list(inputData = "")
            , outputs = list(predictorModel = predictorModel)
            )
        exjson <- epRunGraph(exobj, data = inputData)
    }

    environment(func) <- asNamespace("MicrosoftML")

    ###################################
    ## data.frame
    binaryDf <- within(iris, isSetosa <- Species == "setosa")
    label <- "isSetosa"
    features <- c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")

    predModel <- tempfile("predModel", fileext = ".zip")
    on.exit(if(file.exists(predModel)) unlink(predModel), add = TRUE)

    modelDfOnDisk <- try(func(inputData = binaryDf, label = label, features = features, predictorModel = predModel))
    checkTrue(!inherits(modelDfOnDisk, "try-error"))

    modelDfInMemory <- try(func(inputData = binaryDf, label = label, features = features, predictorModel = ""))
    checkTrue(!inherits(modelDfInMemory, "try-error"))

    ## check
    #modelFileSize <- file.info(predModel)$size
    #checkEquals(modelFileSize, length(modelDfInMemory$outModelBin))

    #outModelBin <- readBin(predModel, what = "raw", n = modelFileSize)
    #checkEquals(outModelBin, modelDfInMemory$outModelBin)
}

"test.ModelOperations.CombineModels2" <- function()
{
    ## use a local function to wrap the non-exported functions
    func <- function(inputData, label, catFeatures, numFeatures, predictorModel)
    {
        transformNode <- Categorical.CatTransformHash(
            Data = "$inputData"
            #, Columns = catFeatures
            , Column = lapply(catFeatures, function(x) list(Source = x))
            , OutputData = "$transformedData"
            , Model = "$transformModel"
            )
        featureNode <- FeatureCombiner.PrepareFeatures(
            Data = "$transformedData"
            , Features = c(catFeatures, numFeatures)
            , OutputData = "$featureData"
            , Model = "$featureModel"
            )
        trainNode <- LogisticRegression.TrainBinary(
            TrainingData = "$featureData"
            , PredictorModel = "$predModel"
            , LabelColumn = label
            , FeatureColumn = "Features"
            )
        combineNode <- ModelOperations.CombineModels(
            TransformModels = c("$transformModel", "$featureModel")
            ,  PredictorModel = "$predModel"
            ,  Model = "$predictorModel")

        exobj <- epCreateGraph(
            transformNode, featureNode, trainNode, combineNode
            , inputs = list(inputData = "")
            , outputs = list(predictorModel = predictorModel)
            )

        exjson <- epRunGraph(exobj, data = inputData)
    }

    ## set the function environment so that non-exported functions can be found
    environment(func) <- asNamespace("MicrosoftML")

    ## prepare data
    claimsXdf <- file.path(rxGetOption("sampleDataDir"), "claims.xdf")
    binaryDf <- rxDataStep(claimsXdf, transforms = list(isTypeA = type == "A"))
    binaryDf <- within(binaryDf, {age <- as.character(age); car.age <- as.character(car.age)})

    label <- "isTypeA"
    catFeatures <- c("age", "car.age")
    numFeatures <- c("cost", "number")

    ## get model in memory
    modelInMemory <- try(func(inputData = binaryDf, label, catFeatures, numFeatures, predictorModel = ""))
    checkTrue(!inherits(modelInMemory, "try-error"))
   
    ## write model to disk
    predModel <- tempfile("predModel", fileext = ".zip")
    on.exit(if(file.exists(predModel)) unlink(predModel), add = TRUE)
    modelOnDisk <- try(func(inputData = binaryDf, label, catFeatures, numFeatures, predictorModel = predModel))
    checkTrue(!inherits(modelOnDisk, "try-error"))

    ## check
    #modelFileSize <- file.info(predModel)$size
    #checkEquals(modelFileSize, length(modelInMemory$outModelBin))

    ###TODO: why are the bytes not always the same?
    #outModelBin <- readBin(predModel, what = "raw", n = modelFileSize)
    #checkEquals(outModelBin, modelInMemory$outModelBin)
}

"test.EnsembleCreator.CreateBinaryEnsemble" <- function()
{
    ## use a local function to wrap the non-exported functions
    func <- function(inputData, label, catFeatures, numFeatures, predictorModel)
    {
        transformNode <- Categorical.CatTransformHash(
            Data = "$inputData"
            #, Columns = catFeatures
            , Column = lapply(catFeatures, function(x) list(Source = x))
            , OutputData = "$transformedData"
            , Model = "$transformModel"
            )
        featureNode <- FeatureCombiner.PrepareFeatures(
            Data = "$transformedData"
            , Features = c(catFeatures, numFeatures)
            , OutputData = "$featureData"
            , Model = "$featureModel"
            )
        trainNode <- LogisticRegression.TrainBinary(
            TrainingData = "$featureData"
            , PredictorModel = "$predModel"
            , LabelColumn = label
            , FeatureColumn = "Features"
            )
        combineNode <- ModelOperations.CombineModels(
            TransformModels = c("$transformModel", "$featureModel")
            ,  PredictorModel = "$predModel"
            ,  Model = "$predictorModel")

        exobj <- epCreateGraph(
            transformNode, featureNode, trainNode, combineNode
            , inputs = list(inputData = "")
            , outputs = list(predictorModel = predictorModel)
            )

        exjson <- epRunGraph(exobj, data = inputData)
    }

    ## set the function environment so that non-exported functions can be found
    environment(func) <- asNamespace("MicrosoftML")

    ## prepare data
    claimsXdf <- file.path(rxGetOption("sampleDataDir"), "claims.xdf")
    binaryDf <- rxDataStep(claimsXdf, transforms = list(isTypeA = type == "A"))
    binaryDf <- within(binaryDf, {age <- as.character(age); car.age <- as.character(car.age)})

    label <- "isTypeA"
    catFeatures <- c("age", "car.age")
    numFeatures <- c("cost", "number")

    ## get models in memory
    baseModels <- list()

    baseModels[[1]] <- try(func(inputData = binaryDf, label, catFeatures, numFeatures, predictorModel = ""))
    checkTrue(!inherits(baseModels[[1]], "try-error"))
   
    baseModels[[2]] <- try(func(inputData = binaryDf, label, catFeatures, numFeatures, predictorModel = ""))
    checkTrue(!inherits(baseModels[[2]], "try-error"))

    ###########################################################################
    ## Combine the baseModels using entry point
    combFunc <- function(models, predictorModel)
    {
        nmodels <- length(models)

        modelsVar <- paste0("$model", seq.int(nmodels))
        combineNode <- EnsembleCreator.CreateBinaryEnsemble(
            Models = modelsVar
            , ValidatePipelines = FALSE
            , PredictorModel = "$predictorModel"
            )

        exobj <- epCreateGraph(
            combineNode
            , inputs = setNames(rep(list(""), nmodels), gsub("^[$]", "", modelsVar))
            , outputs = list(predictorModel = predictorModel))

        ensemble <- epRunGraph(exobj, inModelsBin = data.frame(value = I(lapply(models, "[[", "outModelBin"))))
    }

    ## set the function environment so that non-exported functions can be found
    environment(combFunc) <- asNamespace("MicrosoftML")

    ensemble <- try(combFunc(models = baseModels, predictorModel = ""))
    checkTrue(!inherits(ensemble, "try-error"))

    ###########################################################################
    ## predict with the ensemble model
    predFunc <- function(inputData, predictorModel)
    {
        trainThreads <- 4
        seed <- 123456789

        ## score separately
        scoreNode <- ScoreModel.Score(
            Data = "$inputData", 
            PredictorModel = "$predictorModel", 
            ScoredData = "$scoredVectorData")
        selectNode <- ScoreModel.SelectColumns(
            Data = "$scoredVectorData"
            , OutputData = "$scoredData")

        exobj <- epCreateGraph(
            scoreNode, selectNode
            , inputs = list(inputData = "", predictorModel = "")
            , outputs = list(scoredData = "")
            )
        exjson <- epRunGraph(exobj, seed = seed, parallel = trainThreads
            , data = inputData, outDataFrame = TRUE, inModelBin = predictorModel)
    }

    ## set the function environment so that non-exported functions can be found
    environment(predFunc) <- asNamespace("MicrosoftML")

    ## check
    ensembleScores <- try(predFunc(inputData = binaryDf, predictorModel = ensemble$outModelBin))
    checkTrue(!inherits(ensembleScores, "try-error"))

    baseScores <- try(predFunc(inputData = binaryDf, predictorModel = baseModels[[1]]$outModelBin))
    checkTrue(!inherits(baseScores, "try-error"))

    checkIdentical(ensembleScores, baseScores)
}

"test.ModelOperations.CombineTransformModels" <- function()
{
    ## use a local function to wrap the non-exported functions
    func <- function(inputData, label, catFeatures, numFeatures, predictorModel)
    {
        transformNode <- Categorical.CatTransformHash(
            Data = "$inputData"
            #, Columns = catFeatures
            , Column = lapply(catFeatures, function(x) list(Source = x))
            , OutputData = "$transformedData"
            , Model = "$transformModel"
            )
        featureNode <- FeatureCombiner.PrepareFeatures(
            Data = "$transformedData"
            , Features = c(catFeatures, numFeatures)
            , OutputData = "$featureData"
            , Model = "$featureModel"
            )
        trainNode <- LogisticRegression.TrainBinary(
            TrainingData = "$featureData"
            , PredictorModel = "$predModel"
            , LabelColumn = label
            , FeatureColumn = "Features"
            )
        combxformNode <- ModelOperations.CombineTransformModels(
            Models = c("$transformModel", "$featureModel")
            , OutputModel = "$dataModel")
        combineNode <- ModelOperations.CombineModels(
            TransformModels = "$dataModel"
            ,  PredictorModel = "$predModel"
            ,  Model = "$predictorModel")

        exobj <- epCreateGraph(
            transformNode, featureNode, trainNode, combxformNode, combineNode
            , inputs = list(inputData = "")
            , outputs = list(predictorModel = predictorModel)
            )

        exjson <- epRunGraph(exobj, data = inputData)
    }

    ## set the function environment so that non-exported functions can be found
    environment(func) <- asNamespace("MicrosoftML")

    ## prepare data
    claimsXdf <- file.path(rxGetOption("sampleDataDir"), "claims.xdf")
    binaryDf <- rxDataStep(claimsXdf, transforms = list(isTypeA = type == "A"))
    binaryDf <- within(binaryDf, {age <- as.character(age); car.age <- as.character(car.age)})

    label <- "isTypeA"
    catFeatures <- c("age", "car.age")
    numFeatures <- c("cost", "number")

    ## get model in memory
    modelInMemory <- try(func(inputData = binaryDf, label, catFeatures, numFeatures, predictorModel = ""))
    checkTrue(!inherits(modelInMemory, "try-error"))
   
    ## write model to disk
    predModel <- tempfile("predModel", fileext = ".zip")
    on.exit(if(file.exists(predModel)) unlink(predModel), add = TRUE)
    modelOnDisk <- try(func(inputData = binaryDf, label, catFeatures, numFeatures, predictorModel = predModel))
    checkTrue(!inherits(modelOnDisk, "try-error"))

    ## check
    #modelFileSize <- file.info(predModel)$size
    #checkEquals(modelFileSize, length(modelInMemory$outModelBin))

    ###TODO: why are the bytes not always the same?
    #outModelBin <- readBin(predModel, what = "raw", n = modelFileSize)
    #checkEquals(outModelBin, modelInMemory$outModelBin)
}

"test.ScoreModel.Score" <- function()
{
    func <- function(inputData, label, features, predictorModel)
    {
        trainThreads <- 4
        seed <- 123456789

        featureNode <- FeatureCombiner.PrepareFeatures(
            Data = "$inputData"
            , Features = features
            , OutputData = "$featureData"
            , Model = "$dataModel"
            )
        trainNode <- LogisticRegression.TrainBinary(
            TrainingData = "$featureData"
            , PredictorModel = "$predModel"
            , LabelColumn = label
            , FeatureColumn = "Features"
            , NumThreads = trainThreads
            )
        combNode <- ModelOperations.CombineModels( #Error: *** Exception: 'Unsupported input type: TransformModel'
            TransformModels = "$dataModel"
        #combNode <- ModelOperations.CombineTwoModels(
            #TransformModel = "$dataModel"
            ,  PredictorModel = "$predModel"
            ,  Model = "$predictorModel")

        exobj <- epCreateGraph(
            featureNode, trainNode, combNode
            , inputs = list(inputData = "")
            , outputs = list(predictorModel = predictorModel)
            )
        exjson <- epRunGraph(exobj, seed = seed, parallel = trainThreads, data = inputData)

        ## score separately
        scoreNode <- ScoreModel.Score(
            Data = "$inputData", 
            PredictorModel = "$predictorModel", 
            ScoredData = "$scoredVectorData")
        selectNode <- ScoreModel.SelectColumns(
            Data = "$scoredVectorData"
            , OutputData = "$scoredData")

        exobj <- epCreateGraph(
            scoreNode, selectNode
            , inputs = list(inputData = "", predictorModel = predictorModel)
            , outputs = list(scoredData = "")
            )
        exjson <- epRunGraph(exobj, seed = seed, parallel = trainThreads, data = inputData, outDataFrame = TRUE
            , inModelBin = if(nzchar(predictorModel)) NULL else exjson$outModelBin)
    }

    environment(func) <- asNamespace("MicrosoftML")

    ###################################
    ## data.frame
    binaryDf <- within(iris, isSetosa <- Species == "setosa")
    label <- "isSetosa"
    features <- c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")

    predModel <- tempfile("predModel", fileext = ".zip")
    on.exit(if(file.exists(predModel)) unlink(predModel), add = TRUE)

    modelDfOnDisk <- try(func(inputData = binaryDf, label = label, features = features, predictorModel = predModel))
    checkTrue(!inherits(modelDfOnDisk, "try-error"))

    modelDfInMemory <- try(func(inputData = binaryDf, label = label, features = features, predictorModel = ""))
    checkTrue(!inherits(modelDfInMemory, "try-error"))

    ## check
    #checkEquals(modelDfOnDisk, modelDfInMemory)
}

"test.ScoreModel.SelectColumns" <- function()
{
    func <- function(inputData, label, features, predictorModel)
    {
        featureNode <- FeatureCombiner.PrepareFeatures(
            Data = "$inputData"
            , Features = features
            , OutputData = "$featureData"
            , Model = "$dataModel"
            )
        trainNode <- LogisticRegression.TrainBinary(
            TrainingData = "$featureData"
            , PredictorModel = "$predModel"
            , LabelColumn = label
            , FeatureColumn = "Features"
            )
        combNode <- ModelOperations.CombineModels( #Error: *** Exception: 'Unsupported input type: TransformModel'
            TransformModels = "$dataModel"
        #combNode <- ModelOperations.CombineTwoModels(
            #TransformModel = "$dataModel"
            ,  PredictorModel = "$predModel"
            ,  Model = "$predictorModel")
        scoreNode <- ScoreModel.Score(
            Data = "$inputData"
            , PredictorModel = "$predictorModel"
            , ScoredData = "$scoredVectorData"
            , Suffix = "suffix")
        selectNode <- ScoreModel.SelectColumns(
            Data = "$scoredVectorData"
            , OutputData = "$scoredData"
            , ExtraColumns = c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")
            )

        exobj <- epCreateGraph(
            featureNode, trainNode, combNode, scoreNode, selectNode
            , inputs = list(inputData = "")
            , outputs = list(scoredData = "")
            )
        exjson <- epRunGraph(exobj, data = inputData, outDataFrame = TRUE)
    }

    environment(func) <- asNamespace("MicrosoftML")

    ###################################
    ## data.frame
    binaryDf <- within(iris, isSetosa <- Species == "setosa")
    label <- "isSetosa"
    features <- c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")

    predModel <- tempfile("predModel", fileext = ".zip")
    on.exit(if(file.exists(predModel)) unlink(predModel), add = TRUE)

    modelDfOnDisk <- try(func(inputData = binaryDf, label = label, features = features, predictorModel = predModel))
    checkTrue(!inherits(modelDfOnDisk, "try-error"))

    modelDfInMemory <- try(func(inputData = binaryDf, label = label, features = features, predictorModel = ""))
    checkTrue(!inherits(modelDfInMemory, "try-error"))

    ## check
    #checkEquals(modelDfOnDisk, modelDfInMemory)
}

"test.Categorical.CatTransformDict.TextLoader" <- function()
{
    ## use a local function to wrap the non-exported functions
    func <- function(inputFile, customSchema, predictorModel)
    {
        importNode <- ImportTextData.ImportText(
            InputFile = "$inputFile"
            , Data = "$importedData"
            , CustomSchema = customSchema
            )
        transformNode <- Categorical.CatTransformDict(
            Data = "$importedData"
            #, Columns = "Features"
            , Column = list(list(Source = "Features"))
            , OutputData = "$transformedData"
            )
        trainNode <- LogisticRegression.TrainBinary(
            TrainingData = "$transformedData"
            , PredictorModel = "$predictorModel"
            )

        exobj <- epCreateGraph(
            importNode, transformNode, trainNode
            , inputs = list(inputFile = inputFile)
            , outputs = list(predictorModel = predictorModel)
            )

        exjson <- epRunGraph(exobj)
    }

    ## set the function environment so that non-exported functions can be found
    environment(func) <- asNamespace("MicrosoftML")

    ## prepare data
    claimsXdf <- file.path(rxGetOption("sampleDataDir"), "claims.xdf")
    binaryDf <- rxDataStep(claimsXdf, transforms = list(isTypeA = type == "A"))

    colInfo <- rxCreateColInfo(binaryDf)
    binaryFile <- tempfile("binaryFile", fileext = ".txt")
    binaryText <- RxTextData(binaryFile, colInfo = colInfo)

    rxDataStep(binaryDf, binaryText)
    on.exit(if(file.exists(binaryFile)) unlink(binaryFile), add = TRUE)

    #file.edit(binaryFile)
    customSchema = "separator=, hasHeader=+ column=Label:BL:6 column=Features:TX:1-2"
     
    ## get model in memory
    modelInMemory <- try(func(inputFile = binaryFile, customSchema = customSchema, predictorModel = ""))
    checkTrue(!inherits(modelInMemory, "try-error"))
   
    ## write model to disk
    predModel <- tempfile("predModel", fileext = ".zip")
    on.exit(if(file.exists(predModel)) unlink(predModel), add = TRUE)
    modelOnDisk <- try(func(inputFile = binaryFile, customSchema = customSchema, predictorModel = predModel))
    checkTrue(!inherits(modelOnDisk, "try-error"))

    ## check
    #modelFileSize <- file.info(predModel)$size
    #checkEquals(modelFileSize, length(modelInMemory$outModelBin))

    ###TODO: why are the bytes not always the same?
    #outModelBin <- readBin(predModel, what = "raw", n = modelFileSize)
    #checkEquals(outModelBin, modelInMemory$outModelBin)
}

"test.Categorical.CatTransformHash.TextLoader" <- function()
{
    ## use a local function to wrap the non-exported functions
    func <- function(inputFile, customSchema, predictorModel)
    {
        importNode <- ImportTextData.ImportText(
            InputFile = "$inputFile"
            , Data = "$importedData"
            , CustomSchema = customSchema
            )
        transformNode <- Categorical.CatTransformHash(
            Data = "$importedData"
            #, Columns = "CatFeatures"
            , Column = list(list(Source = "CatFeatures"))
            , OutputData = "$transformedData"
            )
        combineNode <- FeatureCombiner.PrepareFeatures(
            Data = "$transformedData"
            , Features = c("CatFeatures", "NumFeatures")
            , OutputData = "$featureData"
            )
        trainNode <- LogisticRegression.TrainBinary(
            TrainingData = "$featureData"
            , PredictorModel = "$predictorModel"
            )

        exobj <- epCreateGraph(
            importNode, transformNode, combineNode, trainNode
            , inputs = list(inputFile = inputFile)
            , outputs = list(predictorModel = predictorModel)
            )

        exjson <- epRunGraph(exobj)
    }

    ## set the function environment so that non-exported functions can be found
    environment(func) <- asNamespace("MicrosoftML")

    ## prepare data
    claimsXdf <- file.path(rxGetOption("sampleDataDir"), "claims.xdf")
    binaryDf <- rxDataStep(claimsXdf, transforms = list(isTypeA = type == "A"))

    colInfo <- rxCreateColInfo(binaryDf)
    binaryFile <- tempfile("binaryFile", fileext = ".txt")
    binaryText <- RxTextData(binaryFile, colInfo = colInfo)

    rxDataStep(binaryDf, binaryText)
    on.exit(if(file.exists(binaryFile)) unlink(binaryFile), add = TRUE)

    #file.edit(binaryFile)
    customSchema = "separator=, hasHeader=+ column=Label:BL:6 column=CatFeatures:TX:1-2 column=NumFeatures:R4:4-5"
     
    ## get model in memory
    modelInMemory <- try(func(inputFile = binaryFile, customSchema = customSchema, predictorModel = ""))
    checkTrue(!inherits(modelInMemory, "try-error"))
   
    ## write model to disk
    predModel <- tempfile("predModel", fileext = ".zip")
    on.exit(if(file.exists(predModel)) unlink(predModel), add = TRUE)
    modelOnDisk <- try(func(inputFile = binaryFile, customSchema = customSchema, predictorModel = predModel))
    checkTrue(!inherits(modelOnDisk, "try-error"))

    ## check
    #modelFileSize <- file.info(predModel)$size
    #checkEquals(modelFileSize, length(modelInMemory$outModelBin))

    ###TODO: why are the bytes not always the same?
    #outModelBin <- readBin(predModel, what = "raw", n = modelFileSize)
    #checkEquals(outModelBin, modelInMemory$outModelBin)
}

"test.Categorical.CatTransformHash" <- function()
{
    ## use a local function to wrap the non-exported functions
    func <- function(inputData, label, catFeatures, numFeatures, predictorModel)
    {
        transformNode <- Categorical.CatTransformHash(
            Data = "$inputData"
            #, Columns = catFeatures
            , Column = lapply(catFeatures, function(x) list(Source = x))
            , OutputData = "$transformedData"
            )
        combineNode <- FeatureCombiner.PrepareFeatures(
            Data = "$transformedData"
            , Features = c(catFeatures, numFeatures)
            , OutputData = "$featureData"
            )
        trainNode <- LogisticRegression.TrainBinary(
            TrainingData = "$featureData"
            , PredictorModel = "$predictorModel"
            , LabelColumn = label
            , FeatureColumn = "Features"
            )

        exobj <- epCreateGraph(
            transformNode, combineNode, trainNode
            , inputs = list(inputData = "")
            , outputs = list(predictorModel = predictorModel)
            )

        exjson <- epRunGraph(exobj, data = inputData)
    }

    ## set the function environment so that non-exported functions can be found
    environment(func) <- asNamespace("MicrosoftML")

    ## prepare data
    claimsXdf <- file.path(rxGetOption("sampleDataDir"), "claims.xdf")
    binaryDf <- rxDataStep(claimsXdf, transforms = list(isTypeA = type == "A"))
    binaryDf <- within(binaryDf, {age <- as.character(age); car.age <- as.character(car.age)})

    label <- "isTypeA"
    catFeatures <- c("age", "car.age")
    numFeatures <- c("cost", "number")

    ## get model in memory
    modelInMemory <- try(func(inputData = binaryDf, label, catFeatures, numFeatures, predictorModel = ""))
    checkTrue(!inherits(modelInMemory, "try-error"))
   
    ## write model to disk
    predModel <- tempfile("predModel", fileext = ".zip")
    on.exit(if(file.exists(predModel)) unlink(predModel), add = TRUE)
    modelOnDisk <- try(func(inputData = binaryDf, label, catFeatures, numFeatures, predictorModel = predModel))
    checkTrue(!inherits(modelOnDisk, "try-error"))

    ## check
    #modelFileSize <- file.info(predModel)$size
    #checkEquals(modelFileSize, length(modelInMemory$outModelBin))

    ###TODO: why are the bytes not always the same?
    #outModelBin <- readBin(predModel, what = "raw", n = modelFileSize)
    #checkEquals(outModelBin, modelInMemory$outModelBin)
}

"test.Cache.CacheData" <- function()
{
    ## use a local function to wrap the non-exported functions
    func <- function(inputData, label, catFeatures, numFeatures, predictorModel)
    {
        transformNode <- Categorical.CatTransformHash(
            Data = "$inputData"
            #, Columns = catFeatures
            , Column = lapply(catFeatures, function(x) list(Source = x))
            , OutputData = "$transformedData"
            )
        combineNode <- FeatureCombiner.PrepareFeatures(
            Data = "$transformedData"
            , Features = c(catFeatures, numFeatures)
            , OutputData = "$featureData"
            )
        cacheNode <- Cache.CacheData(
            Data = "$featureData"
            , Caching = "Memory"
            , OutputData = "$cachedData"
            )
        trainNode <- LogisticRegression.TrainBinary(
            TrainingData = "$cachedData"
            , PredictorModel = "$predictorModel"
            , LabelColumn = label
            , FeatureColumn = "Features"
            )

        exobj <- epCreateGraph(
            transformNode, combineNode, cacheNode, trainNode
            , inputs = list(inputData = "")
            , outputs = list(predictorModel = predictorModel)
            )

        exjson <- epRunGraph(exobj, data = inputData)
    }

    ## set the function environment so that non-exported functions can be found
    environment(func) <- asNamespace("MicrosoftML")

    ## prepare data
    claimsXdf <- file.path(rxGetOption("sampleDataDir"), "claims.xdf")
    binaryDf <- rxDataStep(claimsXdf, transforms = list(isTypeA = type == "A"))
    binaryDf <- within(binaryDf, {age <- as.character(age); car.age <- as.character(car.age)})

    label <- "isTypeA"
    catFeatures <- c("age", "car.age")
    numFeatures <- c("cost", "number")

    ## get model in memory
    modelInMemory <- try(func(inputData = binaryDf, label, catFeatures, numFeatures, predictorModel = ""))
    checkTrue(!inherits(modelInMemory, "try-error"))
   
    ## write model to disk
    predModel <- tempfile("predModel", fileext = ".zip")
    on.exit(if(file.exists(predModel)) unlink(predModel), add = TRUE)
    modelOnDisk <- try(func(inputData = binaryDf, label, catFeatures, numFeatures, predictorModel = predModel))
    checkTrue(!inherits(modelOnDisk, "try-error"))

    ## check
    #modelFileSize <- file.info(predModel)$size
    #checkEquals(modelFileSize, length(modelInMemory$outModelBin))

    ###TODO: why are the bytes not always the same?
    #outModelBin <- readBin(predModel, what = "raw", n = modelFileSize)
    #checkEquals(outModelBin, modelInMemory$outModelBin)
}

"test.SchemaManipulation.ConcatColumns" <- function()
{
    ## use a local function to wrap the non-exported functions
    func <- function(inputData, label, catFeatures, numFeatures, predictorModel)
    {
        ##TODO: can any entrypoint take two input datasets so that we can concat catFeatures and numFeatures and combine them?
        ## Is the graph only a linear graph?
        concatNode <- SchemaManipulation.ConcatColumns(
            Data = "$inputData"
            #, Columns = catFeatures
            #, OutputColumn = "CatFeatures"
            , Column = list(list(Source = catFeatures, Name = "CatFeatures"))
            , OutputData = "$concatenatedData"
            )
        transformNode <- Categorical.CatTransformHash(
            Data = "$concatenatedData"
            #, Columns = "CatFeatures"
            , Column = list(list(Source = "CatFeatures"))
            , OutputData = "$transformedData"
            )
        combineNode <- FeatureCombiner.PrepareFeatures(
            Data = "$transformedData"
            , Features = c("CatFeatures", numFeatures)
            , OutputData = "$featureData"
            )
        trainNode <- LogisticRegression.TrainBinary(
            TrainingData = "$featureData"
            , PredictorModel = "$predictorModel"
            , LabelColumn = label
            , FeatureColumn = "Features"
            )

        exobj <- epCreateGraph(
            concatNode, transformNode, combineNode, trainNode
            , inputs = list(inputData = "")
            , outputs = list(predictorModel = predictorModel)
            )

        exjson <- epRunGraph(exobj, data = inputData)
    }

    ## set the function environment so that non-exported functions can be found
    environment(func) <- asNamespace("MicrosoftML")

    ## prepare data
    claimsXdf <- file.path(rxGetOption("sampleDataDir"), "claims.xdf")
    binaryDf <- rxDataStep(claimsXdf, transforms = list(isTypeA = type == "A"))
    binaryDf <- within(binaryDf, {age <- as.character(age); car.age <- as.character(car.age)})

    label <- "isTypeA"
    catFeatures <- c("age", "car.age")
    numFeatures <- c("cost", "number")

    ## get model in memory
    modelInMemory <- try(func(inputData = binaryDf, label, catFeatures, numFeatures, predictorModel = ""))
    checkTrue(!inherits(modelInMemory, "try-error"))
   
    ## write model to disk
    predModel <- tempfile("predModel", fileext = ".zip")
    on.exit(if(file.exists(predModel)) unlink(predModel), add = TRUE)
    modelOnDisk <- try(func(inputData = binaryDf, label, catFeatures, numFeatures, predictorModel = predModel))
    checkTrue(!inherits(modelOnDisk, "try-error"))

    ## check
    #modelFileSize <- file.info(predModel)$size
    #checkEquals(modelFileSize, length(modelInMemory$outModelBin))

    ###TODO: why are the bytes not always the same?
    #outModelBin <- readBin(predModel, what = "raw", n = modelFileSize)
    #checkEquals(outModelBin, modelInMemory$outModelBin)
}

"test.SchemaManipulation.SelectColumns" <- function()
{
    ## use a local function to wrap the non-exported functions
    func <- function(inputData, label, catFeatures, numFeatures, predictorModel)
    {
        ##TODO: can any entrypoint take two input datasets so that we can concat catFeatures and numFeatures and combine them?
        ## Is the graph only a linear graph?
        selectNode <- SchemaManipulation.SelectColumns(
            Data = "$inputData"
            #, Columns = c(label, catFeatures, numFeatures)
            , Column = c(label, catFeatures, numFeatures)
            , OutputData = "$selectedData"
            )
        concatNode <- SchemaManipulation.ConcatColumns(
            Data = "$selectedData"
            #, Columns = catFeatures
            #, OutputColumn = "CatFeatures"
            , Column = list(list(Source = catFeatures, Name = "CatFeatures"))
            , OutputData = "$concatenatedData"
            )
        transformNode <- Categorical.CatTransformHash(
            Data = "$concatenatedData"
            #, Columns = "CatFeatures"
            , Column = list(list(Source = "CatFeatures"))
            , OutputData = "$transformedData"
            )
        combineNode <- FeatureCombiner.PrepareFeatures(
            Data = "$transformedData"
            , Features = c("CatFeatures", numFeatures)
            , OutputData = "$featureData"
            )
        trainNode <- LogisticRegression.TrainBinary(
            TrainingData = "$featureData"
            , PredictorModel = "$predictorModel"
            , LabelColumn = label
            , FeatureColumn = "Features"
            )

        exobj <- epCreateGraph(
            selectNode, concatNode, transformNode, combineNode, trainNode
            , inputs = list(inputData = "")
            , outputs = list(predictorModel = predictorModel)
            )

        exjson <- epRunGraph(exobj, data = inputData)
    }

    ## set the function environment so that non-exported functions can be found
    environment(func) <- asNamespace("MicrosoftML")

    ## prepare data
    claimsXdf <- file.path(rxGetOption("sampleDataDir"), "claims.xdf")
    binaryDf <- rxDataStep(claimsXdf, transforms = list(isTypeA = type == "A"))
    binaryDf <- within(binaryDf, {age <- as.character(age); car.age <- as.character(car.age)})

    label <- "isTypeA"
    catFeatures <- c("age", "car.age")
    numFeatures <- c("cost", "number")

    ## get model in memory
    modelInMemory <- try(func(inputData = binaryDf, label, catFeatures, numFeatures, predictorModel = ""))
    checkTrue(!inherits(modelInMemory, "try-error"))
   
    ## write model to disk
    predModel <- tempfile("predModel", fileext = ".zip")
    on.exit(if(file.exists(predModel)) unlink(predModel), add = TRUE)
    modelOnDisk <- try(func(inputData = binaryDf, label, catFeatures, numFeatures, predictorModel = predModel))
    checkTrue(!inherits(modelOnDisk, "try-error"))

    ## check
    #modelFileSize <- file.info(predModel)$size
    #checkEquals(modelFileSize, length(modelInMemory$outModelBin))

    ###TODO: why are the bytes not always the same?
    #outModelBin <- readBin(predModel, what = "raw", n = modelFileSize)
    #checkEquals(outModelBin, modelInMemory$outModelBin)
}

"test.Transform.Only" <- function()
{
    ## use a local function to wrap the non-exported functions
    func <- function(inputData, label, catFeatures, numFeatures, outputData, outVecCol = FALSE)
    {
        ##TODO: can any entrypoint take two input datasets so that we can concat catFeatures and numFeatures and combine them?
        ## Is the graph only a linear graph?
        selectNode <- SchemaManipulation.SelectColumns(
            Data = "$inputData"
            #, Columns = c(label, catFeatures, numFeatures)
            , Column = c(label, catFeatures, numFeatures)
            , OutputData = "$selectedData"
            , Model = "$selectedModel"
            )
        concatNode <- SchemaManipulation.ConcatColumns(
            Data = "$selectedData"
            #, Columns = catFeatures
            #, OutputColumn = "CatFeatures"
            , Column = list(list(Source = catFeatures, Name = "CatFeatures"))
            , OutputData = "$concatenatedData"
            , Model = "$concatenatedModel"
            )
        transformNode <- Categorical.CatTransformHash(
        #transformNode <- Categorical.CatTransformDict(
            Data = "$concatenatedData"
            #, Columns = "CatFeatures"
            , Column = list(list(Source = "CatFeatures"))
            , OutputData = "$transformedData"
            , Model = "$transformedModel"
            )
        featureNode <- FeatureCombiner.PrepareFeatures(
            Data = "$transformedData"
            , Features = c("CatFeatures", numFeatures)
            , OutputData = "$featureData"
            , Model = "$featureModel"
            )

        columns <- c(label, catFeatures, numFeatures)
        if (outVecCol) columns <- c(columns, "CatFeatures")
        selectNode2 <- SchemaManipulation.SelectColumns(
            Data = "$featureData"
            #, Columns = columns
            , Column = columns
            , OutputData = "$outputData"
            , Model = "$selectModel2"
            )

        combxformNode <- ModelOperations.CombineTransformModels(
            Models = c("$selectedModel", "$concatenatedModel", "$transformedModel", "$featureModel", "$selectModel2")
            , OutputModel = "$dataModel")

        outputDataTlc <- if (length(grep("^outputDataTlc", basename(outputData))) > 0) outputData else ""
        exobj <- epCreateGraph(
            selectNode, concatNode, transformNode, featureNode, selectNode2, combxformNode
            , inputs = list(inputData = "")
            , outputs = list(outputData = outputDataTlc, dataModel = "dataModel.zip")   #TODO: can we send the dataModel back to MML by setting dataModel = ""
            )

        outData <- if (nzchar(outputData)) outputData else NULL
        exjson <- epRunGraph(exobj, data = inputData, outData = outData, outDataFrame = is.null(outData), overwrite = TRUE)
    }

    ## set the function environment so that non-exported functions can be found
    environment(func) <- asNamespace("MicrosoftML")

    ## prepare data
    claimsXdf <- file.path(rxGetOption("sampleDataDir"), "claims.xdf")
    binaryDf <- rxDataStep(claimsXdf, transforms = list(isTypeA = type == "A"))
    binaryDf <- within(binaryDf, {age <- as.character(age); car.age <- as.character(car.age)})

    binaryXdf <- tempfile("binaryXdf", fileext = ".xdf")
    on.exit(if(file.exists(binaryXdf)) unlink(binaryXdf), add = TRUE)
    rxImport(inData = binaryDf, outFile = binaryXdf, stringsAsFactors = FALSE)

    label <- "isTypeA"
    catFeatures <- c("age", "car.age")
    numFeatures <- c("cost", "number")

    ## get data in memory
    dataInMemory <- try(func(inputData = binaryDf, label, catFeatures, numFeatures, outputData = ""))
    checkTrue(!inherits(dataInMemory, "try-error"))
   
    dataInMemory <- try(func(inputData = binaryXdf, label, catFeatures, numFeatures, outputData = ""))
    checkTrue(!inherits(dataInMemory, "try-error"))

    ## write data to disk
    outputData <- tempfile("outputData", fileext = ".xdf") #".txt" or ".xdf"
    on.exit(if(file.exists(outputData)) unlink(outputData), add = TRUE)

    dataOnDisk <- try(func(inputData = binaryDf, label, catFeatures, numFeatures, outputData = outputData))
    checkTrue(!inherits(dataOnDisk, "try-error"))
    #head(dataOnDisk)

    dataOnDisk <- try(func(inputData = binaryXdf, label, catFeatures, numFeatures, outputData = outputData))
    checkTrue(!inherits(dataOnDisk, "try-error"))
    #head(dataOnDisk)

    ## write data to disk on TLC side
    outputDataTlc <- tempfile("outputDataTlc", fileext = ".txt") #".txt" or ".idv"
    on.exit(if(file.exists(outputDataTlc)) unlink(outputDataTlc), add = TRUE)

    dataOnDisk <- try(func(inputData = binaryDf, label, catFeatures, numFeatures, outputData = outputDataTlc, outVecCol = TRUE))
    checkTrue(!inherits(dataOnDisk, "try-error"))
    #file.show(dataOnDisk@file)

    dataOnDisk <- try(func(inputData = binaryXdf, label, catFeatures, numFeatures, outputData = outputDataTlc, outVecCol = TRUE))
    checkTrue(!inherits(dataOnDisk, "try-error"))
    #file.show(dataOnDisk@file)
}

"test.SchemaManipulation.DropColumns" <- function()
{
    ## use a local function to wrap the non-exported functions
    func <- function(inputData, label, catFeatures, numFeatures, allFeatures, predictorModel)
    {
        ##TODO: can any entrypoint take two input datasets so that we can concat catFeatures and numFeatures and combine them?
        ## Is the graph only a linear graph?
        dropNode <- SchemaManipulation.DropColumns(
            Data = "$inputData"
            #, Columns = setdiff(allFeatures, c(label, catFeatures, numFeatures))
            , Column = setdiff(allFeatures, c(label, catFeatures, numFeatures))
            , OutputData = "$droppedData"
            )
        concatNode <- SchemaManipulation.ConcatColumns(
            Data = "$droppedData"
            #, Columns = catFeatures
            #, OutputColumn = "CatFeatures"
            , Column = list(list(Source = catFeatures, Name = "CatFeatures"))
            , OutputData = "$concatenatedData"
            )
        transformNode <- Categorical.CatTransformHash(
            Data = "$concatenatedData"
            #, Columns = "CatFeatures"
            , Column = list(list(Source = "CatFeatures"))
            , OutputData = "$transformedData"
            )
        combineNode <- FeatureCombiner.PrepareFeatures(
            Data = "$transformedData"
            , Features = c("CatFeatures", numFeatures)
            , OutputData = "$featureData"
            )
        trainNode <- LogisticRegression.TrainBinary(
            TrainingData = "$featureData"
            , PredictorModel = "$predictorModel"
            , LabelColumn = label
            , FeatureColumn = "Features"
            )

        exobj <- epCreateGraph(
            dropNode, concatNode, transformNode, combineNode, trainNode
            , inputs = list(inputData = "")
            , outputs = list(predictorModel = predictorModel)
            )

        exjson <- epRunGraph(exobj, data = inputData)
    }

    ## set the function environment so that non-exported functions can be found
    environment(func) <- asNamespace("MicrosoftML")

    ## prepare data
    claimsXdf <- file.path(rxGetOption("sampleDataDir"), "claims.xdf")
    binaryDf <- rxDataStep(claimsXdf, transforms = list(isTypeA = type == "A"))
    binaryDf <- within(binaryDf, {age <- as.character(age); car.age <- as.character(car.age)})

    label <- "isTypeA"
    catFeatures <- c("age", "car.age")
    numFeatures <- c("cost", "number")
    allFeatures <- colnames(binaryDf)

    ## get model in memory
    modelInMemory <- try(func(inputData = binaryDf, label, catFeatures, numFeatures, allFeatures, predictorModel = ""))
    checkTrue(!inherits(modelInMemory, "try-error"))
   
    ## write model to disk
    predModel <- tempfile("predModel", fileext = ".zip")
    on.exit(if(file.exists(predModel)) unlink(predModel), add = TRUE)
    modelOnDisk <- try(func(inputData = binaryDf, label, catFeatures, numFeatures, allFeatures, predictorModel = predModel))
    checkTrue(!inherits(modelOnDisk, "try-error"))

    ## check
    #modelFileSize <- file.info(predModel)$size
    #checkEquals(modelFileSize, length(modelInMemory$outModelBin))

    ###TODO: why are the bytes not always the same?
    #outModelBin <- readBin(predModel, what = "raw", n = modelFileSize)
    #checkEquals(outModelBin, modelInMemory$outModelBin)
}

"test.SchemaManipulation.CopyColumns" <- function()
{
    func <- function(inputData, label, features, predictorModel)
    {
        featureNode <- FeatureCombiner.PrepareFeatures(
            Data = "$inputData"
            , Features = features
            , OutputData = "$featureData"
            )
        copyNode <- SchemaManipulation.CopyColumns(
            Data = "$featureData"
            #, Columns = setNames(object = list("CopiedFeatures", "CopiedLabel"), nm = c("Features", label))
            , Column = list(list(Name = "CopiedFeatures", Source = "Features"), list(Name = "CopiedLabel", Source = label))
            , OutputData = "$copiedData"
            )
        trainNode <- LogisticRegression.TrainBinary(
            TrainingData = "$copiedData"
            , PredictorModel = "$predictorModel"
            , LabelColumn = label
            , FeatureColumn = "Features"
            )

        exobj <- epCreateGraph(
            featureNode, copyNode, trainNode
            , inputs = list(inputData = "")
            , outputs = list(predictorModel = predictorModel)
            )

        exjson <- epRunGraph(exobj, data = inputData)
    }

    environment(func) <- asNamespace("MicrosoftML")

    ###################################
    ## data.frame
    binaryDf <- within(iris, isSetosa <- Species == "setosa")
    label <- "isSetosa"
    features <- c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")

    predModel <- tempfile("predModel", fileext = ".zip")
    on.exit(if(file.exists(predModel)) unlink(predModel), add = TRUE)

    modelDfOnDisk <- try(func(inputData = binaryDf, label = label, features = features, predictorModel = predModel))
    checkTrue(!inherits(modelDfOnDisk, "try-error"))

    modelDfInMemory <- try(func(inputData = binaryDf, label = label, features = features, predictorModel = ""))
    checkTrue(!inherits(modelDfInMemory, "try-error"))

    ## check
    #modelFileSize <- file.info(predModel)$size
    #checkEquals(modelFileSize, length(modelDfInMemory$outModelBin))

    #outModelBin <- readBin(predModel, what = "raw", n = modelFileSize)
    #checkEquals(outModelBin, modelDfInMemory$outModelBin)
}
