# Copyright (c) Microsoft Corporation.  All rights reserved.

# Tests for verification of NVidia GPU working.
"test.rxNeuralNet.GPU.enabled" <- function(nnModelOutput, CUDAverificationString) 
{
    DEACTIVATED("Until GPU build machine is available")

    #Gets time ratio between SSE and GPU enabled runs
    GetTimeRatio <- function(CPUrun, GPUrun) 
    {
        timeRatio <- GetTime(CPUrun) / GetTime(GPUrun)
        return(timeRatio)
    }

    #Formats and returns elapsed time value in the rxNeuralNet stdout
    GetTime <- function(nnRunData) 
    {
        for (line in nnRunData) 
        {
            if (grepl("Elapsed time", line))
                elapsedTime <- line
        }

        elapsedTimeValue <- substr(elapsedTime, 15, 30)
        formattedElapedTime <- strptime(elapsedTimeValue, format = "%H:%M:%S")
        formattedElapedTime <- as.character(formattedElapedTime)
        outTime <- substr(formattedElapedTime, 15, 19)
        outTime <- sapply(strsplit(outTime, ":"), ConvertTimeTofloat)
        return(outTime)
    }

    #Converts string represantation of time (min:secs) to float
    ConvertTimeTofloat <- function(x) 
    {
        x <- as.numeric(x)
        x[1] + x[2] / 60
    }

    #Validation that CUDA is enabled on the machine and was used 
    CUDAverificationString <- "Using: GPU CUDA Math on" 
    # This file contains the Net# definition for the neural network structure.
    # More information about the Net# language here: 
    # https://azure.microsoft.com/en-us/documentation/articles/machine-learning-azure-ml-netsharp-reference-guide/
    netDefFile <- system.file("demoScripts/mnist.nn", package = "MicrosoftML")
    source(system.file("extdata/mnist.R", package = "MicrosoftML"))
    mnist <- getMnistData(download = TRUE, sampleDataDir = NULL, createDir = TRUE)
    mnistTrain <- mnist$mnistTrain
    rm(mnist)

    #GPU run output caputre
    nnModelOutput <- capture.output(rxNeuralNet(Label ~ .,
        data = mnistTrain,
        numIterations = 10,
        acceleration = "gpu",
        miniBatchSize = 10,
        normalize = "no",
        optimizer = sgd(learningRate = 0.05, lRateRedRatio = 0.9, lRateRedFreq = 5),
        netDefinition = readChar(netDefFile, file.info(netDefFile)$size),
        type = "multi"))

    #CPU run output capture
    nnModelOutputCPU <- capture.output(rxNeuralNet(Label ~ .,
        data = mnistTrain,
        numIterations = 10,
        acceleration = "sse",
        normalize = "no",
        optimizer = sgd(learningRate = 0.05, lRateRedRatio = 0.9, lRateRedFreq = 5),
        netDefinition = readChar(netDefFile, file.info(netDefFile)$size),
        type = "multi"))

    ratio <- GetTimeRatio(nnModelOutputCPU, nnModelOutput)
   
    CheckStringsFound(nnModelOutput, CUDAverificationString)
    checkTrue(ratio > 2.0)
}

# Helper function for unit tests.
"CheckStringsFound" <- function(outputList, stringsToFind) {
    outputConcatenated <- paste(outputList, collapse = "")
    for (string in stringsToFind) {
        checkTrue(length(grep(string, outputConcatenated)) > 0)
    }
}
