# Copyright (c) Microsoft Corporation.  All rights reserved.

# Tests for basic functionality of rxTlcBridge

"test.rxTlcBridge.code" <- function()
{
    # Create temporary data sources
    myIris = iris
    myIris$Setosa <- iris$Species == "setosa"
    myXdf <- tempfile(pattern = "myIris", fileext = ".xdf")
    on.exit(file.remove(myXdf), add = TRUE)
    rxDataStep(myIris, myXdf, rowsPerRead = 50, overwrite = TRUE)
    myXdfDS <- RxXdfData(file = myXdf)
    
    myCsv <- tempfile(pattern = "myIris", fileext = ".csv")
    on.exit(file.remove(myCsv), add = TRUE)
    myCsvDS <- RxTextData(file = myCsv)
    rxDataStep(myIris, myCsvDS, overwrite = TRUE)
    
   
    myCode <-
	    "tr=lr{l1Weight=0 optTol=1E-05}"


    readVars <- c("Setosa", "Sepal.Length", "Sepal.Width",  "Petal.Width") 
    res1 <- rxTlcBridge(formula = Setosa~Sepal.Length + Sepal.Width + Petal.Width,
        data = myXdfDS, mamlCode = myCode, analysisType = "Train",
	    verbose = 3)
    checkEquals(res1$params$readVars, readVars)
    
    res2 <- rxTlcBridge(formula = Setosa~Sepal.Length + Sepal.Width + Petal.Width,
        data = myIris, mamlCode = myCode, analysisType = "Train")
    checkEquals(res2$params$readVars, readVars)
    
    res3 <- rxTlcBridge(formula = Setosa~Sepal.Length + Sepal.Width + Petal.Width,
        data = myCsvDS, mamlCode = myCode, analysisType = "Train")
    checkEquals(res3$params$readVars, readVars)  
   
}

"test.rxIsExpressEdition.isExpress" <- function()
{
    if (.Platform$OS!="windows")
        DEACTIVATED("Only windows platfrom has express edition.")
    myInfert <- infert
    myInfert$isCase = myInfert$case == 1
    myFormula <- isCase ~ age + parity + spontaneous + induced
	
	if (rxIsExpressEdition()) {
        express.out <- capture.output(rxLogisticRegression(formula = myFormula, 
			data = myInfert, l1 = 0, l2 = 0, normalize = "no", trainThreads = 8))
        cat("\nexpress.out\n");
        print(express.out)
        CheckStringsFound(express.out, "Warning: The number of threads specified in trainer arguments is larger than the concurrency factor setting of the environment. Using 2 training threads instead.")
        #checkTrue(TRUE)
    } else {
        enterprise.out <- capture.output(rxLogisticRegression(formula = myFormula, 
			data = myInfert, l1 = 0, l2 = 0, normalize = "no", trainThreads = 8))
        cat("\nenterprise.out\n");
        print(enterprise.out)
        CheckStringsFound(enterprise.out, "turn off multi-threading by setting trainThreads to 1.")

		rxOptions(isExpress = TRUE)
		checkTrue(rxIsExpressEdition())
		on.exit(rxOptions(isExpress = FALSE), add = TRUE)
		
        express.out <- capture.output(rxLogisticRegression(formula = myFormula, 
			data = myInfert, l1 = 0, l2 = 0, normalize = "no", trainThreads = 8))
        cat("\nexpress.out\n"); print(express.out)
		#checkTrue(length(grep("Warning: The number of threads specified in trainer arguments is larger than the concurrency factor setting of the environment. Using 2 training threads instead.", express.out)) > 0)
        checkTrue(TRUE)
    }
}

# Helper function for unit tests.
"CheckStringsFound" <- function(outputList, stringsToFind) {
    outputConcatenated <- paste(outputList, collapse = "")
    for (string in stringsToFind) {
        checkTrue(length(grep(string, outputConcatenated)) > 0)
    }
}
