# Copyright (c) Microsoft Corporation.  All rights reserved.
# Tests for selectFeatures transform.

"test.selectFeatures.Count" <- function() {
    myData <- data.frame(opinion = c(
        "I love it",
        "I love it",
        "Love it",
        "I love it a lot",
        "Really love it",
        "I love it",
        "I love it",
        "love it",
        "I love it a lot",
        "Really love it",
        "I hate it",
        "I hate it",
        "I hate it",
        "hate it",
        "hate and",
        "I hate it and",
        "I hate it and",
        "I hate it and",
        "hate it and",
        "hate and"),
        like = rep(c(TRUE, FALSE), each = 10),
        stringsAsFactors = FALSE)

    outModel1 <- rxLogisticRegression(like ~ opinionCount, data = myData, l1Weight = 0,
        mlTransforms = list(featurizeText(vars = c(opinionCount = "opinion")), selectFeatures("opinionCount", mode = minCount(9))))
    checkEquals(names(coef(outModel1)), c("(Bias)", "hate", "love", "it", "i"))

    outModel2 <- rxLogisticRegression(like ~ opinionCount, data = myData, l1Weight = 0,
        mlTransforms = list(featurizeText(vars = c(opinionCount = "opinion")), selectFeatures("opinionCount", mode = minCount(5))))
    checkEquals(names(coef(outModel2)), c("(Bias)", "hate", "love", "and", "it", "i"))

    outModel3 <- rxLogisticRegression(like ~ opinionCount, data = myData, l1Weight = 0,
        mlTransforms = list(featurizeText(vars = c(opinionCount = "opinion")), selectFeatures("opinionCount", mode = minCount(15))))
    checkEquals(names(coef(outModel3)), c("(Bias)", "it"))
}

"test.selectFeatures.MutualInformation" <- function() {
    myData <- data.frame(opinion = c(
        "I love it",
        "I love it",
        "Love it",
        "I love it a lot",
        "Really love it",
        "I love it",
        "I love it",
        "love it",
        "I love it a lot",
        "Really love it",
        "I hate it",
        "I hate it",
        "I hate it",
        "hate it",
        "hate and",
        "I hate it and",
        "I hate it and",
        "I hate it and",
        "hate it and",
        "hate and"),
        like = rep(c(TRUE, FALSE), each = 10),
        stringsAsFactors = FALSE)

    form <- like ~ opinionCount
    outModel1 <- rxLogisticRegression(form, data = myData, l1Weight = 0,
        mlTransforms = list(featurizeText(vars = c(opinionCount = "opinion")), 
                            selectFeatures(form, mode = mutualInformation(numFeaturesToKeep = 6))))
    checkEquals(names(coef(outModel1)), c("(Bias)", "hate", "love", "and", "a", "it", "i"))

    outModel2 <- rxLogisticRegression(form, data = myData, l1Weight = 0,
        mlTransforms = list(featurizeText(vars = c(opinionCount = "opinion")),
                            selectFeatures(form, mode = mutualInformation(numFeaturesToKeep = 1))))
    checkEquals(names(coef(outModel2)), c("(Bias)", "love"))
}

"test.mutualInformation.integer" <- function()
{
    DEACTIVATED("Bug 14545: mutualInformation failed if either the label or a feature is of integer type.")
    
    ###################################
    ## integer label
    infert1 <- infert
    #infert1$isCase <- (infert1$case == 1) #OK
    #infert1$isCase <- as.factor(infert1$case == 1) #OK
    #infert1$isCase <- as.numeric(infert1$case == 1) #OK
    infert1$isCase <- as.integer(infert1$case == 1) #failed
    
    form <- isCase ~ age + parity + education + spontaneous + induced
    vars <- list(isCase = c("age" , "parity" , "education" , "spontaneous" , "induced"))
    vecs <- c("age" , "parity" , "education" , "spontaneous" , "induced")

    logitModel <- rxLogisticRegression(form, data = infert1, l1Weight = 0, l2Weight = 0
        , mlTransforms = 
            selectFeatures(form, mutualInformation(numFeaturesToKeep = 3))
            #selectFeatures(vars, mutualInformation(numFeaturesToKeep = 3))
            #selectFeatures(form, mode = minCount(count = 3))
            #selectFeatures(vars, mode = minCount(vars, count = 3))
            #selectFeatures(vecs, mode = minCount(count = 3))
        )

    ###################################
    ## integer feature
    infert1$isCase <- (infert1$case == 1) #failed
    #infert1$isCase <- as.factor(infert1$case == 1) #failed
    #infert1$isCase <- as.numeric(infert1$case == 1) #failed
    #infert1$isCase <- as.integer(infert1$case == 1) #failed
    
    form <- isCase ~ age + parity + education + spontaneous + induced + stratum #stratum is integer

    logitModel <- rxLogisticRegression(form, data = infert1
        , mlTransforms = 
            selectFeatures(form, mutualInformation(numFeaturesToKeep = 3))
            #selectFeatures(vars, mutualInformation(numFeaturesToKeep = 3))
            #selectFeatures(form, mode = minCount(count = 3))
            #selectFeatures(vars, mode = minCount(count = 3))
            #selectFeatures(vecs, mode = minCount(count = 3))
        )
    
    checkTrue(TRUE)
}

"test.featureSelection.factors" <- function()
{
    infert1 <- infert
    infert1$isCase <- as.integer(infert1$case == 1)
    form <- isCase ~ age + parity + education + spontaneous + induced

    logitModel <- rxLogisticRegression(form, data = infert1,
        mlTransforms = selectFeatures(form, mutualInformation(numFeaturesToKeep = 3)))

    checkEquals(names(coef(logitModel)), c("(Bias)", "spontaneous", "induced"))
}
